from django.db import models
from django.contrib.auth.models import User
from random import randint
from math import ceil, sqrt
from datetime import datetime, timedelta
# Create your models here.


"""
This module contains all models for the game itself.
It's good design practice in Django to center all application 
logic into the models.py of the application. So you will see a lot of stuff
that might have gotten into views.py. It basically doesn't matter which way as
long as the codes can't be sensibly called from something other than a view.

The initial data is provided in initial_data.json. This file is automatically
loaded during "syncdb".
 
"""
class Game(models.Model):
	""" Holds game information pertaining to each user. """
	user = models.OneToOneField(User, related_name="game", verbose_name="for user")
	coins = models.IntegerField("coin amount")
	location = models.ForeignKey("Market")
	
		
	maps = models.ManyToManyField("Market", related_name="available_to_players", )
	
	net_worth = models.IntegerField(default=1000)
	def __unicode__(self):
		return u"%s at %s, %i coins" % \
				 (self.user.username,self.location.name,self.coins)
	def travel_to(self, market):
		""" Initiate the voyage to "market" """
		a = self.location
		b = market
		if a == b:
			return
		m = self.location.distance_to(b)
		way = WayProgress(game=self, market_a = a, market_b = b,\
			start_time=datetime.now(), end_time=datetime.now()+timedelta(minutes=m))
		
		way.save()
	def calculate_net_worth(self):
		""" Calculates a player's net worth based on the prices at the central
		market. A simple solution, not entirely precise, but a good aproximation.
		"""
		coins = self.coins
		central = Market.objects.get(id=1)
		for bunch in self.feather_bunches.all():
			stock = central.feather_stocks.get(species=bunch.species)
			coins += stock.price_per_feather*bunch.count
		self.net_worth = coins
		self.save()
		
			
	def check_way(self):
		""" Check if player is on his way somewhere. If so, test for completion.
		"""
		try:
			way = self.way_progress
			# The Beauty of datetime objects is that you can compare them directly
			
			if way.end_time < datetime.now():
				self.location = way.market_b
				way.delete()
				self.save()
		except WayProgress.DoesNotExist:
			pass
			
class BirdSpecies(models.Model):
	"""Different species of birds have different feathers. Prices have been
	decoupled to allow the dynamic pricing. Maybe it should also have
	a link to a wikipedia page about the bird!"""
	name = models.CharField(max_length=40)
	def __unicode__(self):
		return u"%s" % self.name
class FeatherBunch(models.Model):
	"""Feathers in possession of one player. Normally there should only be
	one Featherbunch per type for any player. Technically, there is no check 
	preventing that.
	
	But at least as the code is now, this will not be an issue. Before creating
	a FeatherBunch the code checks if there already is one. However, it is technically possible, that two threads of the "buy" view will instantiate one
	FeatherBunch each. But then again, no player should do that anyway. If so he
	only loses coins. 
	"""
	game = models.ForeignKey(Game, related_name = "feather_bunches", verbose_name="for user")
	species = models.ForeignKey(BirdSpecies, related_name="feather_bunches")
	count = models.IntegerField()
	def __unicode__(self):
		return u"%i %s owned by %s" %(self.count,self.species.name, self.game.user.username)	
	
class Market(models.Model):
	""" Market places have a name and a position. The position
	is used to compute the travel time between two markets."""
	name = models.CharField("name of market",max_length=50)
	position_x = models.IntegerField("x coordinate")
	position_y = models.IntegerField("y coordinate")
	
	map_price = models.IntegerField(default=1000)
	def __unicode__(self):
		return u"%s" % self.name
	
	def distance_to(self,b):
		return sqrt((self.position_x-b.position_x)**2 + (self.position_y-b.position_y)**2)
class WayProgress(models.Model):
	"""If a player is on his way, this object holds information as to when
	the player will arrive.
	"""	
	game = models.OneToOneField(Game, related_name="way_progress")
	market_a = models.ForeignKey(Market, related_name="departures")
	market_b = models.ForeignKey(Market, related_name="arrivals")
	start_time = models.DateTimeField()
	end_time = models.DateTimeField()

class FeatherStock(models.Model):
	"""Price per feather."""
	market = models.ForeignKey(Market, related_name="feather_stocks")
	species = models.ForeignKey(BirdSpecies, related_name="stocks")
	count = models.IntegerField()
	
	max_increment_per_frame = models.IntegerField(default=1)
	max_increment_to = models.IntegerField(default=1)
	increment_probability = models.IntegerField(default=50)
	price_per_feather = models.IntegerField()

	price_variation_min = models.IntegerField(default=1)
	price_variation_max = models.IntegerField(default=10)
	price_variation_target = models.IntegerField(default=1)
	price_variation_reset_probability = models.IntegerField(default=1)
	
	def execute_frame(self):
		"""Execute a frame of the dynamic pricing system. This should be called from manage.py execute_frame or frame_loop."""
		
		if randint(0, 100)< self.increment_probability:
			if self.max_increment_to > self.count:
				self.count += randint(0,self.max_increment_per_frame+1)
				self.save()
				
		if self.price_variation_target != self.price_per_feather:
			change = ceil((self.price_per_feather - self.price_variation_target) / 4)
			if change<0:
				change= min(change, -1)
			else:
				change= max(change, +1)
			print self.price_per_feather, self.price_variation_target
			self.price_per_feather -= change
			self.save()
			
		if randint(0,100)< self.price_variation_reset_probability:
			self.price_variation_target = randint(self.price_variation_min, self.price_variation_max)
			self.save()
	
def execute_frame():
	""" Execute one frame of the game. """
	# Update dynamic pricing 
	for feather_stock in FeatherStock.objects.all():
		feather_stock.execute_frame()
		
	# Calculate HighScores
	for user in User.objects.all():
		try:
			user.game.calculate_net_worth() 
			print user.username, user.game.net_worth
		except Game.DoesNotExist:
			pass
def create_game(user):
	"""Create a game object for a player. Originally thought to hook it up to player registration, but that didn't work out. Now this function gets called
	when the main view throws an exception about the game object not existing.
	"""
	game = Game(user=user, coins=1000,location=Market.objects.get(id=1))
	game.save()
	game.maps.add(Market.objects.get(id=1))
	game.save()
	return game
